Require Import Coqlib Integers Maps.
Require Import Registers.

Require Import Utils.

Require Import MIR DataStructures.

(** Insert a nop at function entry, and ensures all back-edges are nops. *)
Section MIRLOW.
  Definition mapping : Type := PTree.t node.

  Definition back_edge (m_next : mapping * node) dst :=
    let (m, next) := m_next in
      match m ! dst with
        | None => (PTree.set dst next m, Psucc next)
        | Some _ => m_next
      end.

  Definition analyze (f: function) : (mapping * node) :=
    PTree.fold
         (fun (m_next : mapping * node) nd i =>
            match i with
              | Mnop succ
              | Mop _ _ _ succ
              | Mget _ _ _ succ
              | Mput _ _ _ succ
              | Mget_global _ _ succ
              | Mput_global _ _ succ
              | Mnew _ _ succ
              | Mfence succ
              | Mspawn _ _ _ succ
              | Mlock _ succ
              | Munlock _ succ
              | Mcall _ _ _ _ succ
                =>
                  if plt nd succ
                  then m_next
                  else back_edge m_next succ
              | Mif _ _ s1 s2 =>
                  let m_next' := if plt nd s1 then m_next else back_edge m_next s1 in
                    if plt nd s2 then m_next' else back_edge m_next' s2
              | Mret _ => m_next
              end
         )
         f.(fn_code)
         (PTree.empty _, Psucc (max_key f.(fn_code)))
  .

  Definition get_new_dest (m: mapping) (nd: node) : node :=
    match m ! nd with
      | Some nd' => nd'
      | None => nd
    end.

  Definition reroute (m: mapping) : code -> code :=
    let f := get_new_dest m in
    PTree.map
         (fun _ i =>
            match i with
              | Mnop s => Mnop (f s)
              | Mop a b c s => Mop a b c (f s)
              | Mif a b s s' => Mif a b (f s) (f s')
              | Mget a b c s => Mget a b c (f s)
              | Mput a b c s => Mput a b c (f s)
              | Mget_global a b s => Mget_global a b (f s)
              | Mput_global a b s => Mput_global a b (f s)
              | Mnew a b s => Mnew a b (f s)
              | Mfence s => Mfence (f s)
              | Mspawn a b c s => Mspawn a b c (f s)
              | Mlock a s => Mlock a (f s)
              | Munlock a s => Munlock a (f s)
              | Mcall a b c d s => Mcall a b c d (f s)
              | Mret a => Mret a
            end).

  Definition add_nops (m: mapping) : code -> code :=
    PTree.fold
         (fun c dst kk =>
            PTree.set kk (Mnop dst) c
         )
         m.

  Definition lower_mir (f: function) : function :=
    let (m, next) := analyze f in
    let c := reroute m f.(fn_code) in
    let c := add_nops m c in
      mkfunction
        f.(fn_sig)
        f.(fn_params)
        f.(fn_roots)
        (PTree.set next (Mnop f.(fn_entrypoint)) c)
        next.

End MIRLOW.

