(* -*- coq-load-path: ("." "../compcert/cfrontend" "../compcert/lib" "../compcert/common" "../compcert/backend" "../compcert/x86") -*- *)
Require Import Coqlib Registers Integers Maps.

Require Import RTL Traces RTLsort.
Require Import Simulations TSOsimulation Globalenvs.
Require Errors.

Require Import Utils.

Section CORRECTNESS.

  Lemma check_succ :
    forall f nm,
      compute_nodemap f = Errors.OK nm ->
       forall (pc : positive) (i : instruction),
       f.(fn_code) ! pc = Some i ->
       forall s : node, In s (successors_instr i) -> f.(fn_code) ! s <> None.
  Proof.
    unfold compute_nodemap.
    intros f nm H pc i H0 s H1.
    repeat bi. clarify.
    eby eapply check_successors_correct.
  Qed.

  Lemma check_dom_rev :
    forall f nm,
      compute_nodemap f = Errors.OK nm ->
      forall pc, nm ! pc = None -> f.(fn_code) ! pc = None.
  Proof.
    unfold compute_nodemap.
    intros. repeat bi. clarify.
    case_eq (f.(fn_code) ! pc); trivial.
    intros i Hi.
    pose proof (check_dom_correct _ _ Hbi1 Hi).
    clarify.
  Qed.

  Lemma tr_code_correct :
    forall f nm,
      compute_nodemap f = Errors.OK nm ->
      forall n i n', f.(fn_code) ! n = Some i ->
        nm ! n = Some n' ->
        (tr_code nm f.(fn_code)) ! n' = Some (upd_instr nm i).
  Proof.
    unfold tr_code, compute_nodemap.
    intros f nm Hf n i n'.
    repeat bi. clarify.
    apply PTree_Properties.fold_rec.
    intros m m' a H H0 H1 H2. apply H0; congruence.
    rewrite PTree.gempty. congruence.
    intros m a k v. intros Hk Hv A B C.
    rewrite PTree.gsspec.
    unfold upd_node.
    case_eq (nm ! k).
      intros k' Hk'.
      destruct (peq n' k'); clarify.
        assert (U := check_ran_inj nm Hbi2 Hk' C).
        subst. rewrite PTree.gss in B. now clarify.
      apply A; trivial.
      rewrite PTree.gso in B. assumption.
      intros contra. subst. rewrite Hk' in C. clarify.
    intros Hk'.
    elim (check_dom_correct _ _ Hbi1 Hv). trivial.
  Qed.

  Lemma ep_in_dom :
    forall f nm,
      compute_nodemap f = Errors.OK nm ->
      nm ! (f.(fn_entrypoint)) <> None.
  Proof.
    intros f nm H c.
    pose proof (check_dom_rev _ _ H) as K.
    pose proof (K _ c) as Q.
    unfold compute_nodemap in H.
    repeat bi. clarify. unfold check_ep in *.
    rewrite Q in *.
    clarify.
  Qed.

  Definition genv_rel : genv -> genv -> Prop :=
    (fun x y => Genv.genv_match (fun a b => transf_fundef a = Errors.OK b) y x).

  Section MS.
  Variable sge : genv.
  Variable tge : genv.
  Hypothesis TRANSF : genv_rel sge tge.

  Lemma ge_agree_on_symbols s :
    Genv.find_symbol sge s = Genv.find_symbol tge s.
  Proof. eby destruct TRANSF. Qed.

  Lemma function_ptr_translated:
    forall v f,
      Genv.find_funct_ptr sge v = Some f ->
      exists f', transf_fundef f = Errors.OK f' /\
                 Genv.find_funct_ptr tge v = Some f'.
  Proof.
    intros v f H.
    pose proof (proj2 TRANSF v).
    destruct (Genv.find_funct_ptr sge v);
      destruct (Genv.find_funct_ptr tge v); try done.
    clarify; eexists; intuition eauto.
  Qed.

  Lemma find_function_translated:
    forall ros ls f,
      find_function sge ros ls = Some f ->
      exists f', transf_fundef f = Errors.OK f' /\
        find_function tge ros ls = Some f'.
  Proof.
    intros [l|] ls f; simpl.
    destruct (ls # l); try done; simpl. apply function_ptr_translated.
    rewrite ge_agree_on_symbols; destruct (Genv.find_symbol tge i); try done.
    apply function_ptr_translated; done.
  Qed.

  Inductive match_stack_frame : stackframe -> stackframe -> Prop :=
  | MSF f nm r sp n n' rs :
    compute_nodemap f = Errors.OK nm ->
    nm ! n = Some n' ->
    match_stack_frame (Stackframe r f.(fn_code) sp n rs) (Stackframe r (tr_code nm f.(fn_code)) sp n' rs).

  Inductive match_states : state -> state -> Prop :=
  | MS_s f nm stk stk' sp n n' rs:
    list_forall2 match_stack_frame stk stk' ->
    compute_nodemap f = Errors.OK nm ->
    nm ! n = Some n' ->
    match_states (State stk f.(fn_code) sp n rs) (State stk' (tr_code nm f.(fn_code)) sp n' rs)
  | MS_cs stk stk' f f' args :
    transf_fundef f = Errors.OK f' ->
    list_forall2 match_stack_frame stk stk' ->
    match_states (Callstate stk f args) (Callstate stk' f' args)
  | MS_rs stk stk' v :
    list_forall2 match_stack_frame stk stk' ->
    match_states (Returnstate stk v) (Returnstate stk' v)
  | MS_bs stk stk' s :
    list_forall2 match_stack_frame stk stk' ->
    match_states (Blockedstate stk s) (Blockedstate stk' s)
    .

  Let slts := mklts thread_labels (rtl_step sge).
  Let tlts := mklts thread_labels (rtl_step tge).

  Definition bsim_rel := @bsr _ slts tlts match_states.
  Definition bsim_order := @bsc _ tlts.

  Lemma init_sim_succ:
    forall {p vals tinit},
      rtl_init tge p vals = Some tinit ->
      exists sinit, rtl_init sge p vals = Some sinit /\ match_states sinit tinit.
  Proof.
    intros p vals tinit INIT. unfold rtl_init in *.
    pose proof TRANSF as (MG & MF).
    specialize (MF p).
    destruct (Genv.find_funct_ptr tge p) as [[[]|]|]; try done;
      destruct (Genv.find_funct_ptr sge p) as [[[]|]|]; try done; []; simpl in *.
    repeat bi. clarify.
    destruct beq_nat; try done; [].
    eexists; split; try done; inv INIT; vauto.
    constructor. simpl. repeat (autorw; simpl). reflexivity. vauto.
  Qed.

  Lemma init_sim_fail:
    forall {p vals},
      rtl_init tge p vals = None ->
      rtl_init sge p vals = None.
  Proof.
    intros p vals INIT. unfold rtl_init in *.
    pose proof TRANSF as (MG & MF).
    specialize (MF p).
    destruct (Genv.find_funct_ptr tge p);
    destruct (Genv.find_funct_ptr sge p) as [[[]|]|]; try done.
    unfold transf_fundef in MF.
    repeat bi. clarify.
    simpl in *.
    eby destruct beq_nat.
  Qed.

    Ltac succ X :=
      match goal with
        | [|- ?nm ! ?s = Some _] =>
            unfold upd_node;
            case_eq (nm ! s);[trivial|];
            simpl in *;
            intros ?;
            try (eelim X;[left; reflexivity|];
                 eby intuition)
      end.

    Ltac dexists :=
      match goal with
        | [H: exists _, _ |- _] => destruct H
      end.

  Lemma transf_fundef_sig f f' :
    transf_fundef f = Errors.OK f' ->
    funsig f = funsig f'.
  Proof.
    destruct f; simpl; intros; clarify; bi. clarify.
  Qed.

  Lemma transf_fundef_samekind f f' :
    transf_fundef f = Errors.OK f' ->
    match f with
      | Ast.Internal g => exists g', f' = Ast.Internal g'
      | Ast.External g => f' = Ast.External g
    end.
  Proof.
    destruct f; simpl; intros; clarify. bi. clarify. eexists. reflexivity.
  Qed.

  Lemma transf_step_correct:
    forall s ev s', rtl_step sge s ev s' ->
      forall t (MS: match_states s t),
      exists t', rtl_step tge t ev t' /\ match_states s' t'.
  Proof.
    intros s ev s' STEP t MS.
    inv STEP; inv MS;
    try (
        try (match goal with
          | [H: compute_nodemap _ = Errors.OK _,
             K: (fn_code _) ! _ = Some _,
             L: @PTree.get node _ _ = Some _ |- _] =>
              let X := fresh "X" in
              let Y := fresh "Y" in
              pose proof (tr_code_correct _ _ H _ _ _ K L);
              pose proof (check_succ _ _ H _ _ K) as X;
              pose proof (check_dom_rev _ _ H) as Y
             end);
        try (match goal with
          | [H: transf_fundef _ = _ |- _] =>
              let X := fresh "X" in
              let Y := fresh "Y" in
                pose proof (transf_fundef_sig _ _ H) as X;
                pose proof (transf_fundef_samekind _ _ H) as Y;
                simpl in *; try dexists
        end );
        (match goal with
           | [H: Op.eval_operation _ _ _ _ = _ |- _] =>
               rewrite (Op.eval_operation_preserved ge_agree_on_symbols) in H
           | [H: Op.eval_addressing _ _ _ _ = _ |- _] =>
               rewrite (Op.eval_addressing_preserved ge_agree_on_symbols) in H
           | [H: find_function _ ?ros _ = _ |- _] =>
               let Q := fresh "FD" in
                 destruct (find_function_translated _ _ _ H) as (? & FD & ?);
                 pose proof (transf_fundef_sig _ _ FD)
         end || idtac);
        repeat bi; clarify;
        try (eexists; split; vauto; fail);
        try (inv H2; eexists; split; vauto; fail);
        try (
        eexists; split; [vauto|
                         try (constructor; auto; succ X)]; fail)).
    eexists; split; vauto.
    constructor; auto. constructor; auto. constructor; auto. succ X.
    simpl in *.
    eexists; split; vauto.
    constructor; auto. succ X. eelim X. right. left. reflexivity. intuition.

    eexists. split. vauto. simpl. constructor; auto.
    pose proof (ep_in_dom _ _ Hbi).
    succ X. congruence.

    eexists. split. change f.(fn_stacksize) with ((tr_nodes x0 f).(fn_stacksize)). vauto.
    simpl. constructor; auto.
    pose proof (ep_in_dom _ _ Hbi).
    succ X. congruence.

    inv H2; inv H1.
    eexists; split; vauto.
  Qed.

  End MS.

  Definition match_program p p' : Prop :=
    transf_program p = Errors.OK p'.

  Theorem sim : Sim.sim rtl_sem rtl_sem match_program.
  Proof.
    eapply (TSOsim.sim rtl_sem rtl_sem genv_rel bsim_rel (fun _ => bsim_order));
      unfold match_program; simpl; intros.
    rewrite (Ast.transform_partial_program_main _ _ MP). inv GENVR. edone.
    eapply Genv.exists_matching_genv_and_mem_rev
        with (match_var := fun a b => a = b), INIT.
      eby eapply Ast.transform_partial_program_match.
    destruct (init_sim_succ _ _ GENVR INIT) as (i & ? & ?).
      exists i; split;[|apply fsr_in_bsr]; assumption.
    eby eapply init_sim_fail.
    apply forward_to_backward_sim with (fsc:= fun _ _ => False).
      apply rtl_sem.(TSOmachine.SEM_receptive).
      apply rtl_sem.(TSOmachine.SEM_determinate).
      apply rtl_sem.(TSOmachine.SEM_progress_dec).
      split. vauto.
      simpl. intros s t ev s' STEP MS. right. left.
      destruct (transf_step_correct sge tge GENR s ev s' STEP _ MS) as (t' & HS & MS').
      exists t'. intuition. eapply step_weakstep. intuition.
  Qed.

End CORRECTNESS.

